/*****************************************************************************
64show - A simple 6502 disassembler/D64/T64/SID/Pxx/Basic viewer
Copyright (C) 2006 Peter H. Karlsson

Version 1.0

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Steet, Fifth Floor, Boston, MA  02111-1307  USA
****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <ctype.h>
#include <errno.h>
#include <stdbool.h>

/****************************************************************************/
/* globals */
static FILE *file;
static unsigned char *buffer;

/****************************************************************************/
#define VERSION		"1.0"
#define MAXLEN		255	/* max string length */
/* d64 track offset */
#define TRACKOFFSET	5376
/* d64 lengths */
#define LEN_D64_1	174848
#define LEN_D64_2	175531
#define LEN_D64_3	196608
#define LEN_D64_4	197376
/* d64 */
#define NUM_ENTRIES	8
#define IDX_DIRTRACK	18
#define IDX_DISKNAME	144
#define IDX_DISKID	161
#define IDX_FILENAME	5
#define IDX_FTYPE	2
#define IDX_BLOCKL	30
#define IDX_BLOCKH	31
#define LEN_TOTBLOCKS	664
#define LEN_FNAME	16
#define LEN_DISKID	6
#define LEN_DIRBLOCK	32
/* basic */
#define LEN_INBUF	80	/* just basic v2 for now */
#define SIZE_LINELINK   5
#define QUOTE		34
#define BASIC_LOW	128
#define BASIC_HIGH	202
/* sid */
#define SID_NAME	22
#define SID_AUTHOR	54
#define SID_COPYRIGHT	86
#define SID_SONGS	14
#define SID_START	16
#define SID_INIT	10
#define SID_PLAY	12
/* t64 */
#define IDX_INAME	40
#define IDX_TNAME	16
#define IDX_ENTRIESL	36
#define IDX_ENTRIESH	37
#define IDX_STARTL	2
#define IDX_STARTH	3
#define IDX_ENDL	4
#define IDX_ENDH	5
#define IDX_TDIRBLOCK	64
#define LEN_TDIRBLOCK	32
#define LEN_TNAME	16
#define LEN_INAME	24
/* PC64 */
#define IDX_PNAME	8

/****************************************************************************/
typedef enum {
    NORMAL = 0, DISASM, FORCE
} ProgramMode;

typedef enum {
    BIN = 0, BAS, D64, T64, SID, P64
} FileType;

/* 1541 filetype */
enum {	DEL=0x80, SEQ, PRG, USR, REL,
    DEL_S=0xc0, SEQ_S, PRG_S, USR_S, REL_S };

/* t64 filetype */
enum { FREE=0, NORM=1, SNAPSHOT=3 };

/* Addressing modes */
enum { IMPLIED, IMMEDIATE, ABSOLUTE, ABSOLUTE_X, ABSOLUTE_Y, ZEROPAGE,
    INDIRECT_X, INDIRECT_Y, ZEROPAGE_X, INDIRECT, ZEROPAGE_Y, RELATIVE };


/****************************************************************************/
static void exitError(const char *text)
{
    if( NULL != text )
	perror(text);

    if( NULL != buffer )
	free(buffer);

    if( NULL != file )
	fclose(file);

    exit(EXIT_FAILURE);
}

/****************************************************************************/
static void printHelp(void)
{
    fputs("64show " VERSION " (Compiled: " __DATE__ "/" __TIME__ ")\n"
	    "By Peter H. Karlsson <PeterHKarlsson@gmail.com>\n\n"
	    "Usage: 64show [options] file\n"
	    "Options are:\n"
	    "     -a address    disassemble from address (dec/hex 0xnnnn)\n"
	    "     -f            force disassembly\n"
	    "     -h            this help text\n", stderr);
}

/****************************************************************************/
static char printable(const char c)
{
    return (isprint((int)c)?c:'@');
}

/****************************************************************************/
static void convString(const unsigned char *src, unsigned char *target,
	const int len, const bool tospace)
{
    int i;
    int length;
    unsigned char c;

    /* Sanity check */
    if( (NULL == src) || (NULL == target))
	return;

    length = (len<MAXLEN) ? len : MAXLEN;

    for(i = 0; i<length; i++) {
	c = src[i];
	switch(c) {
	    case 0xab:
	    case 0xad:
	    case 0xae:
	    case 0xb0:
	    case 0xb1:
	    case 0xb2:
	    case 0xb3:
	    case 0xbd:
	    case 0xc5:
	    case 0xc9:
	    case 0xca:
	    case 0xcb:
	    case 0xd2:
	    case 0xd5:
	    case 0xdb:
	    case 0xed:
	    case 0xee:
	    case 0xf0:
	    case 0xfd:
		target[i] = '+';
		break;
	    case 0xc0:
	    case 0xc3:
	    case 0xc4:
	    case 0xc6:
		target[i] = '-';
		break;
	    case 0xc2:
	    case 0xc7:
	    case 0xc8:
	    case 0xdd:
		target[i] = '|';
		break;
	    case 0xa0:
		if( tospace )
		    target[i] = ' ';
		else
		    target[i] = c;
		break;
	    default:
		target[i] = (unsigned char)printable((char)c);
		break;
	}
    }
}

/****************************************************************************/
static int offset(const int track, const int sector)
{
    if( ((track>=1 && track<=17) && (sector>=0 && sector<=20))
	    || ((track>=18 && track<=24) && (sector>=0 && sector<=18))
	    || ((track>=25 && track<=30) && (sector>=0 && sector<=17))
	    || ((track>=31 && track<=40) && (sector>=0 && sector<=16)) )
	return ((track-1) * TRACKOFFSET) + (sector * 256);
    else
	return -1;
}

/****************************************************************************/
static FileType getFiletype(const char *fileName, const long int size)
{
    char name[MAXLEN];
    size_t i;
    size_t length;
    FileType ft;

    if( 3 > (i = strlen(fileName)) )
	return BIN;

    length = (i<MAXLEN) ? i : MAXLEN;

    /* "C64File" = 7 */
    if( 7 > size )
	return BIN;

    memset(name, 0, sizeof(name));

    /* lowercase */
    for(i=0; i<length; i++)
	name[i] = (char)tolower(fileName[i]);

    if(
	    ( 0 == strncmp(&name[length-3], "d64", 3) )
	    &&
	    (
	     ( LEN_D64_1 == size ) ||
	     ( LEN_D64_2 == size ) ||
	     ( LEN_D64_3 == size ) ||
	     ( LEN_D64_4 == size )
	    )
      )
    {
	ft = D64;
    }
    else if(
	    ( 0 == strncmp(&name[length-3], "t64", 3) )
	    &&
	    ( 0 == strncmp((char*)buffer, "C64", 3) )
	   )
    {
	ft = T64;
    }
    else if(
	    ( 0 == strncmp(&name[length-3], "sid", 3) )
	    &&
	    ( 0 == strncmp((char*)buffer, "PSID", 4) )
	   )
    {
	ft = SID;
    }
    else if( /* just basic v2 for now */
	    ( 0x01 == (int)buffer[0] )
	    &&
	    ( 0x08 == (int)buffer[1] )
	   )
    {
	ft = BAS;
    }
    else if(
	    ( 0 == strncmp((char*)buffer, "C64File", 7) )
	    &&
	    (
	     ( 'p' == name[length-3] ) ||   /* pxx */
	     ( 's' == name[length-3] ) ||   /* sxx */
	     ( 'u' == name[length-3] ) ||   /* uxx */
	     ( 'r' == name[length-3] )	    /* rxx */
	    )
	   )
    {
	ft = P64;
    }
    else
    {
	ft = BIN;
    }

    return ft;
}

/****************************************************************************/
static const char *filetype1541(const unsigned char c)
{
    switch( c ) {
	case 0:
	    return "SCR";
	case DEL:
	    return "DEL";
	case DEL_S:
	    return "DEL<";
	case SEQ:
	    return "SEQ";
	case SEQ_S:
	    return "SEQ<";
	case PRG:
	    return "PRG";
	case PRG_S:
	    return "PRG<";
	case USR:
	    return "USR";
	case USR_S:
	    return "USR<";
	case REL:
	    return "REL";
	case REL_S:
	    return "REL<";
	default:
	    return "???";
    }
}

/****************************************************************************/
static void d64(void)
{
    unsigned char filename[MAXLEN];
    unsigned char *fname;
    const char *ft;
    int index;
    int	loop;
    int	i;
    int	track;
    int	sector;
    unsigned int totblocks = 0;
    unsigned int blocks	= 0;
    bool more = true;

    /* Sanity check */
    if(NULL == buffer)
	return;

    /* print disk name and id */
    index = offset(IDX_DIRTRACK, 0);
    memset(filename, 0, sizeof(filename));
    convString(&buffer[index+IDX_DISKNAME], filename, LEN_FNAME, true);
    printf("0 \"%s\"", (char *)filename);

    memset(filename, 0, sizeof(filename));
    convString(&buffer[index+IDX_DISKID], filename, LEN_DISKID, true);
    printf("%s\n", (char *)filename);

    /* first directory block */
    index = offset(IDX_DIRTRACK, 1);
    while( more ) {
	/* get next directory sector */
	track	= (int)buffer[index];
	sector	= (int)buffer[index + 1];

	/* 8 file entries in 1 sector */
	for(loop = 0; loop < NUM_ENTRIES; loop++) {
	    if( buffer[index+IDX_FILENAME] != (unsigned char)'\0' ) {
		blocks = (unsigned int)buffer[index+IDX_BLOCKL] +
		    ((unsigned int)buffer[index+IDX_BLOCKH] << 8);
		totblocks += blocks;
		printf("%-5u", blocks);

		/* print filename */
		memset(filename, 0, sizeof(filename));
		convString(&buffer[index+IDX_FILENAME], filename, LEN_FNAME,
			false);
		/* delete shift-space */
		fname = filename + (LEN_FNAME-1);
		while(*fname == (unsigned char)0xa0)
		    *fname-- = '\0';
		printf("\"%s\" ", (char *)filename);
		/* pad with space */
		i = LEN_FNAME - (int)strlen((char *)filename);
		while(i-- > 0)
		    putchar(' ');

		ft = filetype1541(buffer[index+IDX_FTYPE]);
		printf("%s\n", ft);

		if( (0 == strcmp(ft, "SCR")) || (0 == strcmp(ft, "???")) )
		    totblocks -= blocks;

		fflush(NULL);
	    }
	    /* next file entry */
	    index += LEN_DIRBLOCK;
	}
	index = offset(track, sector);
	if( (track == 0) || (index == -1) )
	    more = false;
    }
    printf("%u BLOCKS FREE.\n", ((unsigned int)LEN_TOTBLOCKS-totblocks));
    fflush(NULL);
}

/****************************************************************************/
static void basic(const long int fileSize, unsigned int address)
{
    int low;
    int high;
    unsigned char input;
    unsigned char quote;
    unsigned int next;
    unsigned int line;
    int index;
    unsigned int i;

    /* CBM Basic 2.0 statements */
    const char *basic_statement[] = {
	"END",	    "FOR",	"NEXT",	    "DATA",
	"INPUT#",   "INPUT",    "DIM",	    "READ",
	"LET",	    "GOTO",	"RUN",	    "IF",
	"RESTORE",  "GOSUB",    "RETURN",   "REM",
	"STOP",	    "ON",	"WAIT",	    "LOAD",
	"SAVE",	    "VERIFY",	"DEF",	    "POKE",
	"PRINT#",   "PRINT",    "CONT",	    "LIST",
	"CLR",	    "CMD",	"SYS",	    "OPEN",
	"CLOSE",    "GET",	"NEW",	    "TAB(",
	"TO",	    "FN",	"SPC(",	    "THEN",
	"NOT",	    "STEP",	"+",	    "-",
	"*",	    "/",	"^",	    "AND",
	"OR",	    ">",	"=",	    "<",
	"SGN",	    "INT",	"ABS",	    "USR",
	"FRE",	    "POS",	"SQR",	    "RND",
	"LOG",	    "EXP",	"COS",	    "SIN",
	"TAN",	    "ATN",	"PEEK",	    "LEN",
	"STR$",	    "VAL",	"ASC",	    "CHR$",
	"LEFT$",    "RIGHT$",	"MID$",	    "UNKNOWN"
    };

    /* Sanity check */
    if(NULL == buffer)
	return;

    index = 2; /* skip loading address */;

    do {
	/* get next line link */
	low = (int)buffer[index++];
	high = (int)buffer[index++];

	if( (low == 0) && (high == 0) ) /* last line */
	    break;

	next = ((unsigned int)(low + (high << 8)) - address);
	/* input buffer overflow
	   basic 2.0 max 80 bytes
	   basic 7.0 max 160 bytes */
	if(next > LEN_INBUF)
	    break;

	/* get line number */
	low = (int)buffer[index++];
	high = (int)buffer[index++];
	line = ((unsigned int)low + (high << 8));
	printf("%u ", line);

	/* print current line */
	quote = 0;
	for(i=0; i<next-SIZE_LINELINK ; i++) {
	    input = buffer[index++];
	    if(input == QUOTE) /* toggle quote mode */
		quote ^= input;
	    if( (quote == 0) &&
		    ((input >= BASIC_LOW) && (input <= BASIC_HIGH)) )
		printf("%s", basic_statement[input-BASIC_LOW]);
	    else
		printf("%c", printable((char)input));
	}

	index++;
	address += next;
	putchar('\n');
	fflush(NULL);
    } while(index < fileSize);
    fflush(NULL);
}

/****************************************************************************/
static void disAsm(const long int fileSize, unsigned int address, int memIndex)
{
    unsigned int    low	    = 0;
    unsigned int    high    = 0;
    unsigned int    opcode  = 0;

    const struct {
	char *mnemonic;
	unsigned int size;
	unsigned int type;
    } mnemonics[] = {
	{"BRK",1,IMPLIED},   {"ORA",2,INDIRECT_X},{"???",1,IMPLIED},   {"???",1,IMPLIED},
	{"???",1,IMPLIED},   {"ORA",2,ZEROPAGE},  {"ASL",2,ZEROPAGE},  {"???",1,IMPLIED},
	{"PHP",1,IMPLIED},   {"ORA",2,IMMEDIATE}, {"ASL",1,IMPLIED},   {"???",1,IMPLIED},
	{"???",1,IMPLIED},   {"ORA",3,ABSOLUTE},  {"ASL",3,ABSOLUTE},  {"???",1,IMPLIED},
	{"BPL",2,RELATIVE},  {"ORA",2,INDIRECT_Y},{"???",1,IMPLIED},   {"???",1,IMPLIED},
	{"???",1,IMPLIED},   {"ORA",2,ZEROPAGE_X},{"ASL",2,ZEROPAGE_X},{"???",1,IMPLIED},
	{"CLC",1,IMPLIED},   {"ORA",3,ABSOLUTE_Y},{"???",1,IMPLIED},   {"???",1,IMPLIED},
	{"???",1,IMPLIED},   {"ORA",3,ABSOLUTE_X},{"ASL",3,ABSOLUTE_X},{"???",1,IMPLIED},
	{"JSR",3,ABSOLUTE},  {"AND",2,INDIRECT_X},{"???",1,IMPLIED},   {"???",1,IMPLIED},
	{"BIT",2,ZEROPAGE},  {"AND",2,ZEROPAGE},  {"ROL",2,ZEROPAGE},  {"???",1,IMPLIED},
	{"PLP",1,IMPLIED},   {"AND",2,IMMEDIATE}, {"ROL",1,IMPLIED},   {"???",1,IMPLIED},
	{"BIT",3,ABSOLUTE},  {"AND",3,ABSOLUTE},  {"ROL",3,ABSOLUTE},  {"???",1,IMPLIED},
	{"BMI",2,RELATIVE},  {"AND",2,INDIRECT_Y},{"???",1,IMPLIED},   {"???",1,IMPLIED},
	{"???",1,IMPLIED},   {"AND",2,ZEROPAGE_X},{"ROL",2,ZEROPAGE_X},{"???",1,IMPLIED},
	{"SEC",1,IMPLIED},   {"AND",3,ABSOLUTE_Y},{"???",1,IMPLIED},   {"???",1,IMPLIED},
	{"???",1,IMPLIED},   {"AND",3,ABSOLUTE_X},{"ROL",3,ABSOLUTE_X},{"???",1,IMPLIED},
	{"RTI",1,IMPLIED},   {"EOR",2,INDIRECT_X},{"???",1,IMPLIED},   {"???",1,IMPLIED},
	{"???",1,IMPLIED},   {"EOR",2,ZEROPAGE},  {"LSR",2,ZEROPAGE},  {"???",1,IMPLIED},
	{"PHA",1,IMPLIED},   {"EOR",2,IMMEDIATE}, {"LSR",1,IMPLIED},   {"???",1,IMPLIED},
	{"JMP",3,ABSOLUTE},  {"EOR",3,ABSOLUTE},  {"LSR",3,ABSOLUTE},  {"???",1,IMPLIED},
	{"BVC",2,RELATIVE},  {"EOR",2,INDIRECT_Y},{"???",1,IMPLIED},   {"???",1,IMPLIED},
	{"???",1,IMPLIED},   {"EOR",2,ZEROPAGE_X},{"LSR",2,ZEROPAGE_X},{"???",1,IMPLIED},
	{"CLI",1,IMPLIED},   {"EOR",3,ABSOLUTE_Y},{"???",1,IMPLIED},   {"???",1,IMPLIED},
	{"???",1,IMPLIED},   {"EOR",3,ABSOLUTE_X},{"???",1,IMPLIED},   {"???",1,IMPLIED},
	{"RTS",1,IMPLIED},   {"ADC",2,INDIRECT_X},{"???",1,IMPLIED},   {"???",1,IMPLIED},
	{"???",1,IMPLIED},   {"ADC",2,ZEROPAGE},  {"ROR",2,ZEROPAGE},  {"???",1,IMPLIED},
	{"PLA",1,IMPLIED},   {"ADC",2,IMMEDIATE}, {"ROR",1,IMPLIED},   {"???",1,IMPLIED},
	{"JMP",3,INDIRECT},  {"ADC",3,ABSOLUTE},  {"ROR",3,ABSOLUTE},  {"???",1,IMPLIED},
	{"BVS",2,RELATIVE},  {"ADC",2,INDIRECT_Y},{"???",1,IMPLIED},   {"???",1,IMPLIED},
	{"???",1,IMPLIED},   {"ADC",2,ZEROPAGE_X},{"ROR",2,ZEROPAGE_X},{"???",1,IMPLIED},
	{"SEI",1,IMPLIED},   {"ADC",3,ABSOLUTE_Y},{"???",1,IMPLIED},   {"???",1,IMPLIED},
	{"???",1,IMPLIED},   {"ADC",3,ABSOLUTE_X},{"ROR",3,ABSOLUTE_X},{"???",1,IMPLIED},
	{"???",1,IMPLIED},   {"STA",2,INDIRECT_X},{"???",1,IMPLIED},   {"???",1,IMPLIED},
	{"STY",2,ZEROPAGE},  {"STA",2,ZEROPAGE},  {"STX",2,ZEROPAGE},  {"???",1,IMPLIED},
	{"DEY",1,IMPLIED},   {"???",1,IMPLIED},   {"TXA",1,IMPLIED},   {"???",1,IMPLIED},
	{"STY",3,ABSOLUTE},  {"STA",3,ABSOLUTE},  {"STX",3,ABSOLUTE},  {"???",1,IMPLIED},
	{"BCC",2,RELATIVE},  {"STA",2,INDIRECT_Y},{"???",1,IMPLIED},   {"???",1,IMPLIED},
	{"STY",2,ZEROPAGE_X},{"STA",2,ZEROPAGE_X},{"???",1,IMPLIED},   {"???",1,IMPLIED},
	{"TYA",1,IMPLIED},   {"STA",3,ABSOLUTE_Y},{"TXS",1,IMPLIED},   {"???",1,IMPLIED},
	{"???",1,IMPLIED},   {"STA",3,ABSOLUTE_X},{"???",1,IMPLIED},   {"???",1,IMPLIED},
	{"LDY",2,IMMEDIATE}, {"LDA",2,INDIRECT_X},{"LDX",2,IMMEDIATE}, {"???",1,IMPLIED},
	{"LDY",2,ZEROPAGE},  {"LDA",2,ZEROPAGE},  {"LDX",2,ZEROPAGE},  {"???",1,IMPLIED},
	{"TAY",1,IMPLIED},   {"LDA",2,IMMEDIATE}, {"TAX",1,IMPLIED},   {"???",1,IMPLIED},
	{"LDY",3,ABSOLUTE},  {"LDA",3,ABSOLUTE},  {"LDX",3,ABSOLUTE},  {"???",1,IMPLIED},
	{"BCS",2,RELATIVE},  {"LDA",2,INDIRECT_Y},{"???",1,IMPLIED},   {"???",1,IMPLIED},
	{"LDY",2,ZEROPAGE_X},{"LDA",2,ZEROPAGE_X},{"LDX",2,ZEROPAGE_Y},{"???",1,IMPLIED},
	{"CLV",1,IMPLIED},   {"LDA",3,ABSOLUTE_Y},{"TSX",1,IMPLIED},   {"???",1,IMPLIED},
	{"LDY",3,ABSOLUTE_X},{"LDA",3,ABSOLUTE_X},{"LDX",3,ABSOLUTE_Y},{"???",1,IMPLIED},
	{"CPY",2,IMMEDIATE}, {"CMP",2,INDIRECT_X},{"???",1,IMPLIED},   {"???",1,IMPLIED},
	{"CPY",2,ZEROPAGE},  {"CMP",2,ZEROPAGE},  {"DEC",2,ZEROPAGE},  {"???",1,IMPLIED},
	{"INY",1,IMPLIED},   {"CMP",2,IMMEDIATE}, {"DEX",1,IMPLIED},   {"???",1,IMPLIED},
	{"CPY",3,ABSOLUTE},  {"CMP",3,ABSOLUTE},  {"DEC",3,ABSOLUTE},  {"???",1,IMPLIED},
	{"BNE",2,RELATIVE},  {"CMP",2,INDIRECT_Y},{"???",1,IMPLIED},   {"???",1,IMPLIED},
	{"???",1,IMPLIED},   {"CMP",2,ZEROPAGE_X},{"DEC",2,ZEROPAGE_X},{"???",1,IMPLIED},
	{"CLD",1,IMPLIED},   {"CMP",3,ABSOLUTE_Y},{"???",1,IMPLIED},   {"???",1,IMPLIED},
	{"???",1,IMPLIED},   {"CMP",3,ABSOLUTE_X},{"DEC",3,ABSOLUTE_X},{"???",1,IMPLIED},
	{"CPX",2,IMMEDIATE}, {"SBC",2,INDIRECT_X},{"???",1,IMPLIED},   {"???",1,IMPLIED},
	{"CPX",2,ZEROPAGE},  {"SBC",2,ZEROPAGE},  {"INC",2,ZEROPAGE},  {"???",1,IMPLIED},
	{"INX",1,IMPLIED},   {"SBC",2,IMMEDIATE}, {"NOP",1,IMPLIED},   {"???",1,IMPLIED},
	{"CPX",3,ABSOLUTE},  {"SBC",3,ABSOLUTE},  {"INC",3,ABSOLUTE},  {"???",1,IMPLIED},
	{"BEQ",2,RELATIVE},  {"SBC",2,INDIRECT_Y},{"???",1,IMPLIED},   {"???",1,IMPLIED},
	{"???",1,IMPLIED},   {"SBC",2,ZEROPAGE_X},{"INC",2,ZEROPAGE_X},{"???",1,IMPLIED},
	{"SED",1,IMPLIED},   {"SBC",3,ABSOLUTE_Y},{"???",1,IMPLIED},   {"???",1,IMPLIED},
	{"???",1,IMPLIED},   {"SBC",3,ABSOLUTE_X},{"INC",3,ABSOLUTE_X},{"???",1,IMPLIED}
    };

    /* Sanity check */
    if(NULL == buffer)
	return;

    while (memIndex < fileSize) {
	opcode = buffer[memIndex++];
	printf("$%04X  ", address);
	switch (mnemonics[opcode].size) {
	    case 2:
		low = buffer[memIndex++];
		address += 2;
		printf("%02X %02X     ", opcode, low);
		break;
	    case 3:
		low = buffer[memIndex++];
		high = buffer[memIndex++];
		address += 3;
		printf("%02X %02X %02X  ", opcode, low, high);
		break;
	    default:
		address++;
		printf("%02X        ", opcode);
		break;
	}

	printf("%s ", mnemonics[opcode].mnemonic);
	switch (mnemonics[opcode].type) {
	    case IMMEDIATE:
		printf("#$%02X", low);
		break;
	    case ABSOLUTE:
		printf("$%02X%02X", high, low);
		break;
	    case ABSOLUTE_X:
		printf("$%02X%02X,X", high, low);
		break;
	    case ABSOLUTE_Y:
		printf("$%02X%02X,Y", high, low);
		break;
	    case ZEROPAGE:
		printf("$%02X", low);
		break;
	    case INDIRECT_X:
		printf("($%02X,X)", low);
		break;
	    case INDIRECT_Y:
		printf("($%02X),Y", low);
		break;
	    case ZEROPAGE_X:
		printf("$%02X,X", low);
		break;
	    case INDIRECT:
		printf("($%02X%02X)", high, low);
		break;
	    case ZEROPAGE_Y:
		printf("$%02X,Y", low);
		break;
	    case RELATIVE:
		printf("$%04X", (low<=127)?(address+low):(address-(256-low)));
		break;
	    default:
		/* IMPLIED */
		break;
	}

	/* empty line efter JMP and RTS */
	switch( opcode ) {
	    case 0x4c:
	    case 0x60:
	    case 0x6c:
		putchar('\n');
		break;
	    default:
		break;
	}
	putchar('\n');
	fflush(NULL);
    }
    fflush(NULL);
}

/****************************************************************************/
static void sid(void)
{
    /* Sanity check */
    if( NULL == buffer )
	return;

    printf("Name:       %s\n", (char*)&buffer[SID_NAME]);
    printf("Author:     %s\n", (char*)&buffer[SID_AUTHOR]);
    printf("Copyright:  %s\n\n", (char*)&buffer[SID_COPYRIGHT]);
    printf("%-6s %-7s %-5s %-5s\n", "Songs", "Default", "Init", "Play");
    printf("==========================\n");
    printf("%-6d ", ((int)buffer[SID_SONGS] << 8) + (int)buffer[SID_SONGS+1]);
    printf("%-7d ", ((int)buffer[SID_START] << 8) + (int)buffer[SID_START+1]);
    printf("$%04x ", ((unsigned int)buffer[SID_INIT] << 8) +
	    (unsigned int)buffer[SID_INIT+1]);
    printf("$%04x\n", ((unsigned int)buffer[SID_PLAY] << 8) +
	    (unsigned int)buffer[SID_PLAY+1]);
    fflush(NULL);
}

/****************************************************************************/
static void t64(void)
{
    char name[MAXLEN];
    int entries;
    int i;
    int index;

    /* Sanity check */
    if(NULL == buffer)
	return;

    memset(name, 0, sizeof(name));
    strncpy(name, (char*)&buffer[IDX_INAME], LEN_INAME);
    printf("Name: \"%s\"\n", name);

    printf("Filename            C64s     1541 Start/End\n");
    printf("=============================================\n");

    entries = (int)(buffer[IDX_ENTRIESH]*256) + (int)buffer[IDX_ENTRIESL];
    if(entries < 1) {
	printf("No files found.\n");
	fflush(NULL);
	return;
    }

    index = IDX_TDIRBLOCK;
    for(i = 0; i<entries; i++) {
	memset(name, 0, sizeof(name));
	convString(&buffer[index+IDX_TNAME], (unsigned char*)name,
		LEN_TNAME, false);
	printf("%-18s  ", name);

	switch(buffer[index]) {
	    case FREE:
		printf("Free     ");
		break;
	    case NORM:
		printf("Normal   ");
		break;
	    case SNAPSHOT:
		printf("Snapshot ");
		break;
	    default:
		printf("Unknown  ");
		break;
	}

	printf("%-5s", filetype1541(buffer[index+1]));

	printf("$%04x-", (unsigned int)buffer[index+IDX_STARTL] +
		((unsigned int)buffer[index+IDX_STARTH] << 8));
	printf("$%04x\n", (unsigned int)buffer[index+IDX_ENDL] +
		((unsigned int)buffer[index+IDX_ENDH] << 8));

	index += LEN_TDIRBLOCK;
    }
    fflush(NULL);
}

/****************************************************************************/
static void p64(void)
{
    unsigned char filename[MAXLEN];

    /* Sanity check */
    if(NULL == buffer)
	return;

    memset(filename, 0, sizeof(filename));

    convString(&buffer[IDX_PNAME], filename,
	    (int)strlen((char*)&buffer[IDX_PNAME]), false);
    printf("%s\n", (char*)filename);
    fflush(NULL);
}

/****************************************************************************/
int main(int argc, char **argv)
{
    long int fileSize;
    int fileIndex = 0;
    int memIndex;
    unsigned int address = 0;
    char *endPtr;
    ProgramMode pMode = NORMAL;

    /* clear globals */
    file = NULL;
    buffer = NULL;

    switch( argc ) {
	case 2:
	    if( 0 == strcmp(argv[1], "-h") ) {
		printHelp();
		return EXIT_SUCCESS;
	    }
	    pMode = NORMAL;
	    fileIndex = 1;
	    break;
	case 3:
	    if( 0 == strcmp(argv[1], "-f") ) {
		pMode = FORCE;
		fileIndex = 2;
	    } else {
		printHelp();
		return EXIT_FAILURE;
	    }
	    break;
	case 4:
	    if( 0 == strcmp(argv[1], "-a") ) {
		address = (unsigned int)strtol(argv[2], &endPtr, 0);
		if ( endPtr == argv[2] ) {
		    errno = EINVAL;
		    exitError("strtol");
		}
		pMode = DISASM;
		fileIndex = 3;
	    } else {
		printHelp();
		return EXIT_FAILURE;
	    }
	    break;
	default:
	    printHelp();
	    return EXIT_FAILURE;
    }

    if(NULL == (file = fopen(argv[fileIndex], "rb")) )
	exitError("fopen()");

    /* get file size */
    if( 0 > fseek(file, 0L, SEEK_END) )
	exitError("fseek()");
    if( 0 > (fileSize = ftell(file)) )
	exitError("ftell()");
    if( 0 > fseek(file, 0L, SEEK_SET) )
	exitError("fseek()");

    if( NULL == (buffer = calloc((size_t)fileSize, sizeof(unsigned char))) )
	exitError("calloc()");

    if( 0 == (fread(buffer, sizeof(unsigned char), (size_t)fileSize, file)) )
	exitError("fread()");

    fclose(file);
    file = NULL;

    if( DISASM != pMode )
	address = (unsigned int)(buffer[0] + (buffer[1] << 8));

    memIndex = 0;

    if( NORMAL == pMode ) {
	switch( getFiletype(argv[fileIndex], fileSize) ) {
	    case D64:
		d64();
		break;
	    case T64:
		t64();
		break;
	    case SID:
		sid();
		break;
	    case P64:
		p64();
		break;
	    case BAS:
		basic(fileSize, address);
		break;
	    case BIN:
	    default:
		memIndex = 2;
		disAsm(fileSize, address, memIndex);
		break;
	}
    } else if ( DISASM == pMode ) {
	disAsm(fileSize, address, memIndex);
    } else if ( FORCE == pMode ) {
	memIndex = 2;
	disAsm(fileSize, address, memIndex);
    }

    free(buffer);
    return EXIT_SUCCESS;
}
